# =============================================================================
# Twitwi Normalizers
# =============================================================================
#
# Function used to "normalize" tweets from various sources (v1 API, v2 API,
# public scraped API etc.) in order to produce homogeneous and easily
# analysable data.
#
import re
from copy import deepcopy
from datetime import datetime
from html import unescape

from twitwi.utils import (
    get_dates,
    custom_normalize_url,
    validate_payload_v2,
    custom_get_normalized_hostname
)

CLEAN_RT_PATTERN = re.compile(r'^RT @\w+: ')


def get_collection_time():
    return datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')


def format_rt_text(user, text):
    return 'RT @%s: %s' % (user, text)


def format_qt_text(user, text, quoted_text, url):
    quote = '« %s: %s — %s »' % (user, quoted_text, url)

    return text.replace(url, quote)


def format_tweet_url(screen_name, tweet_id):
    return 'https://twitter.com/%s/status/%s' % (screen_name, tweet_id)


def extract_media_name_from_url(media_url):
    return media_url.rsplit('/', 1)[-1].split('?tag=', 1)[0]


def extract_items_from_text(text, char):
    splitter = re.compile(r'[^\w%s]+' % char)

    return sorted(
        set(
            r.lstrip(char).lower()
            for r in splitter.split(CLEAN_RT_PATTERN.sub('', text))
            if r.startswith(char)
        )
    )


def extract_hashtags_from_text(text):
    return extract_items_from_text(text, '#')


def extract_mentions_from_text(text):
    return extract_items_from_text(text, '@')


def resolve_entities(tweet, prefix):
    status_key = '%s_status' % prefix
    target = tweet[status_key]

    for ent in ['entities', 'extended_entities']:
        if ent not in target:
            continue
        tweet[ent] = tweet.get(ent, {})
        for field in target[ent]:
            tweet[ent][field] = tweet[ent].get(field, [])
            if field in target[ent]:
                tweet[ent][field] += target[ent][field]


def get_bitrate(x):
    return x.get('bitrate', 0)


def nostr_field(f):
    return f.replace('_str', '')


META_FIELDS = [
    'in_reply_to_status_id_str',
    'in_reply_to_screen_name',
    'in_reply_to_user_id_str',
    'lang',
    'possibly_sensitive',
    'retweet_count',
    'favorite_count',
    'reply_count'
]

META_FIELD_TRANSLATIONS = {
    'in_reply_to_status_id_str': 'to_tweetid',
    'in_reply_to_screen_name': 'to_username',
    'in_reply_to_user_id_str': 'to_userid',
    'favorite_count': 'like_count'
}

USER_META_FIELDS = [
    'screen_name',
    'name',
    'friends_count',
    'followers_count',
    'location',
    'verified',
    'description',
    'created_at'
]

PLACE_META_FIELDS = [
    'country_code',
    'full_name',
    'place_type'
]


def grab_extra_meta(source, result, locale=None):

    if source.get('coordinates'):
        result['coordinates'] = source['coordinates']['coordinates']
        result['lat'] = source['coordinates']['coordinates'][1]
        result['lng'] = source['coordinates']['coordinates'][0]
    else:

        # TODO: this is hardly optimal
        result['coordinates'] = None

    for meta in META_FIELDS:
        if meta in source:
            result[META_FIELD_TRANSLATIONS.get(meta, meta)] = source[meta]
        elif nostr_field(meta) in source:
            result[meta] = str(source[nostr_field(meta)])

    for meta in USER_META_FIELDS:
        key = 'user_%s' % meta.replace('_count', '')
        if key in source:
            result[key] = source[key]
        elif 'user' in source and meta in source['user']:
            result[key] = source['user'][meta] if source['user'][meta] != "" else None

    if 'user' in source:
        result['user_id'] = source['user']['id_str']
        result['user_tweets'] = source['user']['statuses_count']
        result['user_likes'] = source['user']['favourites_count']
        result['user_lists'] = source['user']['listed_count']
        result['user_image'] = source['user']['profile_image_url_https']

    if 'place' in source and source['place'] is not None:
        for meta in PLACE_META_FIELDS:
            if meta in source['place']:
                key = 'place_%s' % meta.replace('place_', '').replace('full_', '')
                result[key] = source['place'][meta]

        if 'bounding_box' in source['place'] \
                and source['place']['bounding_box'] is not None \
                and 'coordinates' in source['place']['bounding_box']:
            result['place_coordinates'] = source['place']['bounding_box']['coordinates'][0]

    # TODO: nested_get
    try:
        result['user_url'] = source['user']['entities']['url']['urls'][0]['expanded_url']
    except (KeyError, IndexError):
        try:
            result['user_url'] = source['user']['url']
        except KeyError:
            pass

    if 'user_created_at' in result:
        result['user_timestamp_utc'], result['user_created_at'] = get_dates(result['user_created_at'], locale)

    if source.get('source'):
        result['source_url'], result['source_name'] = source['source'].replace('<a href="', '').replace('</a>', '').split('" rel="nofollow">')

    return result


def normalize_tweet(tweet, locale=None, extract_referenced_tweets=False,
                    collection_source=None, pure=True):
    """
    Function "normalizing" a tweet as returned by Twitter's API in order to
    cleanup and optimize some fields.

    Note that this function mutates the argument to work.

    Args:
        tweet (dict): Tweet json dict from Twitter API.
        locale (pytz.timezone, optional): Timezone for date conversions.
        extract_referenced_tweets (bool, optional): Whether to return only
            the original tweet or the full list of tweets found in the given
            tweet payload (including quoted and retweeted tweets). Defaults
            to `False`.
        collection_source (str, optional): string explaining how the tweet
            was collected. Defaults to `None`.
        pure (bool, optional): whether to allow the function to mutate its
            original argument. Default to `True`.

    Returns:
        (dict or list): Either a single tweet dict or a list of tweet dicts if
            `extract_referenced_tweets` was set to `True`.

    """

    if pure:
        tweet = deepcopy(tweet)

    results = []

    if 'extended_tweet' in tweet:
        for field in tweet['extended_tweet']:
            tweet[field] = tweet['extended_tweet'][field]

    text = tweet.get('full_text', tweet.get('text', ''))

    rti = None
    rtu = None
    rtuid = None
    rtime = None

    if 'retweeted_status' in tweet and tweet['retweeted_status']['id_str'] != tweet['id_str']:
        rti = tweet['retweeted_status']['id_str']
        rtu = tweet['retweeted_status']['user']['screen_name']
        rtuid = tweet['retweeted_status']['user']['id_str']

        nested = normalize_tweet(
            tweet['retweeted_status'],
            locale=locale,
            extract_referenced_tweets=True,
            collection_source='retweet',
            pure=False
        )

        rtweet = nested[-1]

        if extract_referenced_tweets:
            results.extend(nested)

        rtime = rtweet['timestamp_utc']

        resolve_entities(tweet, 'retweeted')

    qti = None
    qtu = None
    qtuid = None
    qtime = None

    if 'quoted_status' in tweet and tweet['quoted_status']['id_str'] != tweet['id_str']:
        qti = tweet['quoted_status']['id_str']
        qtu = tweet['quoted_status']['user']['screen_name']
        qtuid = tweet['quoted_status']['user']['id_str']

        nested = normalize_tweet(
            tweet['quoted_status'],
            locale=locale,
            extract_referenced_tweets=True,
            collection_source='quote',
            pure=False
        )

        qtweet = nested[-1]

        if extract_referenced_tweets:
            results.extend(nested)

        if 'quoted_status_permalink' in tweet:
            qturl = tweet['quoted_status_permalink']['expanded']
        else:
            qturl = qtweet['url']
        qtime = qtweet['timestamp_utc']

        resolve_entities(tweet, 'quoted')

    medids = set()
    media_urls = []
    media_files = []
    media_types = []
    media_alt_texts = []

    links = set()
    hashtags = set()
    mentions = {}

    if 'entities' in tweet or 'extended_entities' in tweet:
        source_id = rti or qti or tweet['id_str']

        entities = tweet.get('extended_entities', tweet['entities']).get('media', [])
        entities += tweet['entities'].get('urls', [])

        for entity in entities:
            if 'expanded_url' in entity and 'url' in entity and entity['expanded_url']:
                try:
                    text = text.replace(entity['url'], entity['expanded_url'])
                except KeyError:
                    pass

            if 'media_url' in entity:
                if 'video_info' in entity:
                    med_url = max(entity['video_info']['variants'], key=get_bitrate)['url']
                else:
                    med_url = entity['media_url_https']

                med_name = extract_media_name_from_url(med_url)

                if med_name not in medids:
                    medids.add(med_name)
                    media_types.append(entity['type'])
                    media_urls.append(med_url.split('?tag=')[0])
                    media_files.append('%s_%s' % (source_id, med_name))
                    media_alt_texts.append(entity.get("ext_alt_text") or '')
            else:
                normalized = custom_normalize_url(entity['expanded_url'])
                links.add(normalized)

        for hashtag in tweet['entities'].get('hashtags', []):
            hashtags.add(hashtag['text'].lower())

        for mention in tweet['entities'].get('user_mentions', []):
            mentions[mention['screen_name'].lower()] = mention['id_str']

    if rtu:
        text = format_rt_text(rtu, rtweet['text'])

    if qtu:
        text = format_qt_text(qtu, text, qtweet['text'], qturl)

    timestamp_utc, local_time = get_dates(tweet['created_at'], locale)
    text = unescape(text)

    if collection_source is None:
        collection_source = tweet.get('collection_source')
    links = sorted(links)
    domains = [custom_get_normalized_hostname(l, normalize_amp=False, infer_redirection=False) for l in links]
    normalized_tweet = {
        'id': tweet['id_str'],
        'local_time': local_time,
        'timestamp_utc': timestamp_utc,
        'text': text,
        'url': format_tweet_url(tweet['user']['screen_name'], tweet['id_str']),
        'quoted_id': qti,
        'quoted_user': qtu,
        'quoted_user_id': qtuid,
        'quoted_timestamp_utc': qtime,
        'retweeted_id': rti,
        'retweeted_user': rtu,
        'retweeted_user_id': rtuid,
        'retweeted_timestamp_utc': rtime,
        'media_files': media_files,
        'media_types': media_types,
        'media_urls': media_urls,
        'media_alt_texts': media_alt_texts,
        'links': links,
        'links_to_resolve': len(links) > 0,
        'domains': domains,
        'hashtags': sorted(hashtags) if hashtags else extract_hashtags_from_text(text),
        'mentioned_ids': [mentions[m] for m in sorted(mentions.keys())],
        'mentioned_names': sorted(mentions.keys()) if mentions else extract_mentions_from_text(text),
        'collection_time': get_collection_time(),
        'match_query': collection_source != 'thread' and collection_source != 'quote'
    }

    if collection_source is not None:
        normalized_tweet['collected_via'] = [collection_source]

    grab_extra_meta(tweet, normalized_tweet, locale)

    results.append(normalized_tweet)

    if not extract_referenced_tweets:
        return results[0]

    return results


def resolve_user_entities(user):
    if 'entities' in user:
        for k in user['entities']:
            if 'urls' in user['entities'][k]:
                for url in user['entities'][k]['urls']:
                    if not url.get('expanded_url'):
                        continue
                    if k in user:
                        user[k] = user[k].replace(url['url'], url['expanded_url'])


def normalize_user(user, locale=None, pure=True, v2=False):
    """
    Function "normalizing" a user as returned by Twitter's API in order to
    cleanup and optimize some fields.

    Note that this function mutates the argument to work.

    Args:
        user (dict): Twitter user json dict from Twitter API.
        locale (pytz.timezone, optional): Timezone for date conversions.
        pure (bool, optional): whether to allow the function to mutate its
            original argument. Default to `True`.

    Returns:
        dict: The normalized user.

    """

    if pure:
        user = deepcopy(user)

    resolve_user_entities(user)

    timestamp_utc, local_time = get_dates(user['created_at'], locale, v2)

    if v2 and 'withheld' in user:
        withheld = user['withheld']
        withheld_in_countries = withheld.get('country_codes', [])
        withheld_scope = withheld.get('withheld_scope', '')
    else:
        withheld_in_countries = []
        withheld_scope = ''

    normalized_user = {
        'id': user['id_str'] if not v2 else user['id'],
        'screen_name': user['screen_name'] if not v2 else user['username'],
        'name': user['name'],
        'description': user['description'] if user["description"] else None,
        'url': user['url'],
        'timestamp_utc': timestamp_utc,
        'local_time': local_time,
        'location': user.get('location'),
        'verified': user.get('verified'),
        'protected': user.get('protected'),
        'tweets': user['statuses_count'] if not v2 else user['public_metrics']['tweet_count'],
        'followers': user['followers_count'] if not v2 else user['public_metrics']['followers_count'],
        'friends': user['friends_count'] if not v2 else user['public_metrics']['following_count'],
        'likes': user['favourites_count'] if not v2 else None,
        'lists': user['listed_count'] if not v2 else user['public_metrics']['listed_count'],
        'image': user.get('profile_image_url_https') if not v2 else user.get('profile_image_url'),
        'default_profile': user.get('default_profile', False),
        'default_profile_image': user.get('default_profile_image', False),
        'witheld_in_countries': user.get('witheld_in_countries', []) if not v2 else withheld_in_countries,
        'witheld_scope': user.get('witheld_scope') if not v2 else withheld_scope
    }

    return normalized_user


def includes_index(payload, key, index_key='id'):
    return {item[index_key]: item for item in payload['includes'].get(key, [])}


def get_best_url(item):
    if 'unwound_url' in item:
        return item['unwound_url']

    if 'expanded_url' in item:
        return item['expanded_url']

    return None


def normalize_tweet_v2(tweet, *, users_by_screen_name, places_by_id, tweets_by_id,
                       users_by_id, media_by_key, locale=None, collection_source=None,
                       extract_referenced_tweets=False):
    timestamp_utc, local_time = get_dates(tweet['created_at'], locale=locale, v2=True)

    user = users_by_id[tweet['author_id']]
    user_timestamp_utc, user_created_at = get_dates(user['created_at'], locale=locale, v2=True)
    user_entities = user.get('entities', {})

    entities = tweet.get('entities', {})
    referenced_tweets = tweet.get('referenced_tweets', [])

    hashtags = set()

    for hashtag in entities.get('hashtags', []):
        hashtags.add(hashtag['tag'])

    mentions = {}

    for mention in entities.get('mentions', []):
        if 'id' in mention:
            mentions[mention['username']] = mention['id']
        else:
            mentions[mention['username']] = users_by_screen_name[mention['username']]['id']

    place_info = {}

    if 'geo' in tweet:
        geo_data = tweet['geo']

        if 'coordinates' in geo_data:
            point = geo_data['coordinates']

            if point['type'] == 'Point':
                lng, lat = point['coordinates']
                place_info['lng'] = lng
                place_info['lat'] = lat

        if 'place_id' in geo_data:
            place_data = places_by_id.get(geo_data['place_id'], {})

            if 'country_code' in place_data:
                place_info['place_country_code'] = place_data['country_code']

            if 'full_name' in place_data:
                place_info['place_name'] = place_data['full_name']

            if 'place_type' in place_data:
                place_info['place_type'] = place_data['place_type']

            if 'geo' in place_data and 'bbox' in place_data['geo']:
                place_info['place_coordinates'] = place_data['geo']['bbox']

    # Text
    text = tweet['text']

    # References
    refs = {t['type']: t['id'] for t in referenced_tweets}

    # Reply
    reply_info = {}

    if 'replied_to' in refs:
        reply = tweets_by_id.get(refs['replied_to'], {})
        if 'author_id' in reply:
            reply_info['to_username'] = users_by_id[reply['author_id']]['username']
        else:
            reply_info['to_username'] = ''
        reply_info['to_userid'] = reply.get('author_id', '')
        reply_info['to_tweetid'] = reply.get('id', '')

    # Retweet
    retweet_info = {}
    normalized_retweet = None

    if 'retweeted' in refs:
        retweet_info['retweeted_id'] = refs['retweeted']

        if refs['retweeted'] in tweets_by_id:
            retweet = tweets_by_id[refs['retweeted']]
            normalized_retweet = normalize_tweet_v2(
                retweet,
                users_by_screen_name=users_by_screen_name,
                places_by_id=places_by_id,
                tweets_by_id=tweets_by_id,
                users_by_id=users_by_id,
                media_by_key=media_by_key,
                locale=locale,
                collection_source='retweet'
            )

            retweet_info['retweeted_user'] = normalized_retweet['user_screen_name']
            retweet_info['retweeted_user_id'] = normalized_retweet['user_id']
            retweet_info['retweeted_timestamp_utc'] = normalized_retweet['timestamp_utc']

    # Quoted
    quote_info = {}
    normalized_quote = None

    if 'quoted' in refs:
        quote_info['quoted_id'] = refs['quoted']

        if refs['quoted'] in tweets_by_id:
            quote = tweets_by_id[refs['quoted']]
            normalized_quote = normalize_tweet_v2(
                quote,
                users_by_screen_name=users_by_screen_name,
                places_by_id=places_by_id,
                tweets_by_id=tweets_by_id,
                users_by_id=users_by_id,
                media_by_key=media_by_key,
                locale=locale,
                collection_source='quote'
            )

            quote_info['quoted_user'] = normalized_quote['user_screen_name']
            quote_info['quoted_user_id'] = normalized_quote['user_id']
            quote_info['quoted_timestamp_utc'] = normalized_quote['timestamp_utc']

    # Replace urls in text
    links = set()

    for url_data in entities.get('urls', []):
        replacement_url = get_best_url(url_data)

        if replacement_url:
            text = text.replace(url_data['url'], replacement_url)

        if not replacement_url:
            replacement_url = url_data['url']

        links.add(custom_normalize_url(replacement_url))

    if normalized_retweet:
        text = format_rt_text(
            normalized_retweet['user_screen_name'],
            normalized_retweet['text']
        )

    if normalized_quote:
        text = format_qt_text(
            normalized_quote['user_screen_name'],
            text,
            normalized_quote['text'],
            normalized_quote['url']
        )

    # Metrics
    is_retweet = 'retweeted' in refs
    public_metrics = tweet['public_metrics']
    user_public_metrics = user['public_metrics']

    user_url = user['url']

    if 'url' in user_entities and 'urls' in user_entities['url']:
        user_url_entity = user_entities['url']['urls'][0]
        user_url = get_best_url(user_url_entity) or user_url

    medias = []

    if 'attachments' in tweet and 'media_keys' in tweet['attachments']:
        source_id = refs.get('retweeted_id', tweet['id'])

        for media_key in tweet['attachments']['media_keys']:
            if media_key in media_by_key:
                media_data = media_by_key[media_key]

                medias.append((
                    media_data.get('url', ''),
                    '%s_%s' % (source_id, extract_media_name_from_url(media_data.get('url', ''))),
                    media_data['type']
                ))

    if collection_source is None:
        collection_source = tweet.get('collection_source')

    sorted_mentions = sorted(mentions.keys())

    normalized_tweet = {
        'id': tweet['id'],
        'local_time': local_time,
        'timestamp_utc': timestamp_utc,
        'text': unescape(text),
        'url': format_tweet_url(user['username'], tweet['id']),
        'hashtags': sorted(hashtags),
        'mentioned_names': sorted_mentions,
        'mentioned_ids': [mentions[k] for k in sorted_mentions],
        'collection_time': get_collection_time(),
        'user_id': user['id'],
        'user_screen_name': user['username'],
        'user_name': user['name'],
        'user_image': user['profile_image_url'],
        'user_url': user_url or None,
        'user_location': user.get("location"),
        'user_verified': user['verified'],
        'user_description': user['description'] if user["description"] else None,
        'user_tweets': user_public_metrics['tweet_count'],
        'user_followers': user_public_metrics['followers_count'],
        'user_friends': user_public_metrics['following_count'],
        'user_lists': user_public_metrics['listed_count'],
        'user_created_at': user_created_at,
        'user_timestamp_utc': user_timestamp_utc,
        'possibly_sensitive': tweet['possibly_sensitive'],
        'like_count': public_metrics['like_count'] if not is_retweet else 0,
        'retweet_count': public_metrics['retweet_count'] if not is_retweet else 0,
        'quote_count': public_metrics['quote_count'] if not is_retweet else 0,
        'reply_count': public_metrics['reply_count'] if not is_retweet else 0,
        'lang': tweet['lang'],
        'source_name': tweet.get('source'),
        'links': sorted(links),
        'media_urls': [m[0] for m in medias],
        'media_files': [m[1] for m in medias],
        'media_types': [m[2] for m in medias],
        'match_query': collection_source != 'thread' and collection_source != 'quote',
        **place_info,
        **reply_info,
        **retweet_info,
        **quote_info
    }

    if collection_source is not None:
        normalized_tweet['collected_via'] = [collection_source]

    if extract_referenced_tweets:
        normalized_tweets = [normalized_tweet]

        if normalized_retweet is not None:
            normalized_tweets.append(normalized_retweet)

        if normalized_quote is not None:
            normalized_tweets.append(normalized_quote)

        return normalized_tweets

    return normalized_tweet


def normalize_tweets_payload_v2(payload, locale=None, extract_referenced_tweets=False,
                                collection_source=None):

    if not validate_payload_v2(payload):
        raise TypeError('given value is not a Twitter API v2 payload')

    if 'data' not in payload:
        return []

    users_by_screen_name = includes_index(payload, 'users', index_key='username')
    users_by_id = includes_index(payload, 'users')
    places_by_id = includes_index(payload, 'places')
    tweets_by_id = includes_index(payload, 'tweets')
    media_by_key = includes_index(payload, 'media', index_key='media_key')

    output = []
    already_seen = {}

    for item in payload['data']:
        normalized_tweets = normalize_tweet_v2(
            item,
            users_by_id=users_by_id,
            users_by_screen_name=users_by_screen_name,
            places_by_id=places_by_id,
            tweets_by_id=tweets_by_id,
            media_by_key=media_by_key,
            locale=locale,
            collection_source=collection_source,
            extract_referenced_tweets=True
        )

        if extract_referenced_tweets:
            for normalized_tweet in normalized_tweets:
                k = int(normalized_tweet['id'])
                earlier_normalized_tweet = already_seen.get(k)

                if earlier_normalized_tweet is not None:
                    if 'collected_via' in normalized_tweet:
                        new_collection_source = normalized_tweet['collected_via'][0]

                        if 'collected_via' not in earlier_normalized_tweet:
                            earlier_normalized_tweet['collected_via'] = [new_collection_source]
                        else:
                            if new_collection_source not in earlier_normalized_tweet['collected_via']:
                                earlier_normalized_tweet['collected_via'].append(new_collection_source)

                    continue

                already_seen[k] = normalized_tweet
                output.append(normalized_tweet)
        else:
            assert len(normalized_tweets)
            output.append(normalized_tweets[0])

    return output



# =============================================================================
# Twitwi Constants
# =============================================================================
#
# Useful constants used throughout the library.
#
TWEET_DATETIME_FORMAT = '%a %b %d %H:%M:%S +0000 %Y'
TWEET_DATETIME_FORMAT_V2 = '%Y-%m-%dT%H:%M:%S.%fZ'
FORMATTED_TWEET_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

# More details on Twitter's tweets metadata can be read here: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/tweet-object
TWEET_FIELDS = [
    'id',                             # digital ID
    'timestamp_utc',                  # UNIX timestamp of creation - UTC time
    'local_time',                     # ISO datetime of creation - local time
    'user_screen_name',               # author's user text ID (@user) (at collection time)
    'text',                           # message's text content
    # 'filter_level',                   # maximum value of the filter_level parameter which may be used and still stream this Tweet
    'possibly_sensitive',             # whether a link present in the message might contain sensitive content according to Twitter
    # 'withheld_copyright',             # whether the tweet might be censored by Twitter following copyright requests, ignorable
    # 'withheld_scope',                 # whether the content withheld is the 'status' or a 'user', ignorable
    # 'withheld_countries',             # list of ISO country codes in which the message is withheld, separated by |, ignorable
    # 'truncated',                      # whether the tweet is bigger than 140 characters, obsolete
    'retweet_count',                  # number of retweets of the message (at collection time)
    'like_count',                     # number of likes of the message (at collection time)
    'reply_count',                    # number of answers to the message, dropped by Twitter (since Oct 17, now charged), unreliable and ignorable
    'lang',                           # language of the message automatically identified by Twitter's algorithms (equals 'und' when no language could be detected)
    'to_username',                    # text ID of the user the message is answering to
    'to_userid',                      # digital ID of the user the message is answering to
    'to_tweetid',                     # digital ID of the tweet the message is answering to
    # 'source',                         # medium used by the user to post the message, now exported in source_name and source_url fields
    'source_name',                    # name of the medium used to post the message
    'source_url',                     # link to the medium used to post the message
    'user_location',                  # location declared in the user's profile (at collection time)
    'lat',                            # latitude of messages geolocalized
    'lng',                            # longitude of messages geolocalized
    'user_id',                        # author's user digital ID
    'user_name',                      # author's detailed textual name (at collection time)
    'user_verified',                  # whether the author's account is certified
    'user_description',               # description given in the author's profile (at collection time)
    'user_url',                       # link to a website given in the author's profile (at collection time)
    'user_image',                     # link to the image avatar of the author's profile (at collection time)
    # 'user_utcoffset',                 # time offset due to the user's timezone, dropped by Twitter (since May 18), ignorable
    # 'user_timezone',                  # timezone declared in the user's profile, dropped by Twitter (since May 18), ignorable
    # 'user_lang',                      # language declared in the user's profile (at collection time), dropped by Twitter (since May 19), ignorable
    'user_tweets',                    # number of tweets sent by the user (at collection time)
    'user_followers',                 # number of users following the author (at collection time)
    'user_friends',                   # number of users the author is following (at collection time)
    'user_likes',                     # number of likes the author has expressed (at collection time)
    'user_lists',                    # number of users lists the author has been included in (at collection time)
    'user_created_at',                # ISO datetime of creation of the author's account
    'user_timestamp_utc',             # UNIX timestamp of creation of the author's account - UTC time
    'collected_via',                  # How we received the message: 'stream', 'search', 'retweet' (the original tweet was
                                    # contained in the retweet metadata), 'quote' (the original tweet was contained in
                                    # the quote metadata), 'thread' (the tweet is part of the same conversation as a
                                    # tweet collected via search or stream). If the message was collected via multiple
                                    # ways, they are separated by |
    'match_query',                    # whether the tweet was retrieved because it matches the query, or whether it was
                                    # collected via 'quote' or 'thread'
    'retweeted_id',                   # digital ID of the retweeted message
    'retweeted_user',                 # text ID of the user who authored the retweeted message
    'retweeted_user_id',              # digital ID of the user who authoring the retweeted message
    'retweeted_timestamp_utc',        # UNIX timestamp of creation of the retweeted message - UTC time
    'quoted_id',                      # digital ID of the retweeted message
    'quoted_user',                    # text ID of the user who authored the quoted message
    'quoted_user_id',                 # digital ID of the user who authoring the quoted message
    'quoted_timestamp_utc',           # UNIX timestamp of creation of the quoted message - UTC time
    'collection_time',                # ISO datetime of message collection - local time
    'url',                            # url of the tweet (to get a view of the message directly on Twitter)
    'place_country_code',             # if the tweet has an associated 'place', country code of that place
    'place_name',                     # if the tweet has an associated 'place', name of that place
    'place_type',                     # if the tweet has an associated 'place', type of that place ('city', 'admin', etc.)
    'place_coordinates',              # if the tweet has an associated 'place', coordinates of that place, separated by |
    'links',                          # list of links included in the text content, with redirections resolved, separated by |
    'domains',                        # list of domain names in the links fields, separated by |
    'media_urls',                     # list of links to images/videos embedded, separated by |
    'media_files',                    # list of filenames of images/videos embedded and downloaded, separated by |, ignorable when medias collections isn't enabled
    'media_types',                    # list of media types (photo, video, animated gif), separated by |
    'media_alt_texts',                # list of alternative texts (image descriptions), separated by |
    'mentioned_names',                # list of text IDs of users mentionned, separated by |
    'mentioned_ids',                  # list of digital IDs of users mentionned, separated by |
    'hashtags'                        # list of hashtags used, lowercased, separated by |
]

TWEET_FIELDS_TCAT = [
    'id',
    'time',
    'created_at',
    'from_user_name',
    'text',
    'filter_level',
    'possibly_sensitive',
    'withheld_copyright',
    'withheld_scope',
    'truncated',
    'retweet_count',
    'favorite_count',
    'lang',
    'to_user_name',
    'in_reply_to_status_id',
    'quoted_status_id',
    'source',
    'location',
    'lat',
    'lng',
    'from_user_id',
    'from_user_realname',
    'from_user_verified',
    'from_user_description',
    'from_user_url',
    'from_user_profile_image_url',
    'from_user_utcoffset',
    'from_user_timezone',
    'from_user_lang',
    'from_user_tweetcount',
    'from_user_followercount',
    'from_user_friendcount',
    'from_user_favourites_count',
    'from_user_listed',
    'from_user_withheld_scope',
    'from_user_created_at',
    'urls',
    'urls_expanded',
    'urls_followed',
    'domains',
    'HTTP status code',
    'media_id',
    'media_urls',
    'media_type',
    'media_indice_start',
    'media_indice_end',
    'photo_sizes_width',
    'photo_sizes_height',
    'photo_resize',
    'mentions',
    'hashtags'
]

GAZOU_TO_TCAT = {

    'identical_fields': {
        'id': 'id',
        'timestamp_utc': 'time',
        'local_time': 'created_at',
        'user_screen_name': 'from_user_name',
        'text': 'text',
        'possibly_sensitive': 'possibly_sensitive',
        'retweet_count': 'retweet_count',
        'like_count': 'favorite_count',
        'lang': 'lang',
        'to_username': 'to_user_name',
        'to_userid': 'to_user_id',
        'to_tweetid': 'in_reply_to_status_id',
        'quoted_id': 'quoted_status_id',
        'user_location': 'location',
        'lat': 'lat',
        'lng': 'lng',
        'user_id': 'from_user_id',
        'user_name': 'from_user_realname',
        'user_verified': 'from_user_verified',
        'user_description': 'from_user_description',
        'user_url': 'from_user_url',
        'user_image': 'from_user_profile_image_url',
        'user_tweets': 'from_user_tweetcount',
        'user_followers': 'from_user_followercount',
        'user_friends': 'from_user_friendcount',
        'user_likes': 'from_user_favourites_count',
        'user_lists': 'from_user_listed',
        'user_created_at': 'from_user_created_at',
        'links': 'urls_expanded',
        'domains': 'domains',
        'mentioned_ids': 'mentions',
        'hashtags': 'hashtags',
        'media_urls': 'media_urls',
        'media_types': 'media_type'
    },

    'modified_fields': [
        'source'
    ],

    'removed_fields': [
        'filter_level',
        'witheld_copyright',
        'withheld_scope',
        'truncated',
        'from_user_utcoffset',
        'from_user_timezone',
        'from_user_lang',
        'from_user_withheld_scope',
        'urls',
        'media_id',
        'media_indice_start',
        'media_indice_end',
        'photo_sizes_width',
        'photo_sizes_height',
        'photo_resize',
    ]
}

TWEET_PLURAL_FIELDS = {
    'links',
    'urls_expanded',
    'domains',
    'hashtags',
    'collected_via',
    'media_urls',
    'media_files',
    'media_types',
    'media_alt_texts',
    'mentioned_names',
    'mentioned_ids',
    'mentions'
}

TWEET_BOOLEAN_FIELDS = {
    'possibly_sensitive',
    'user_verified',
    'match_query'
}

# More details on Twitter's users metadata can be read here: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/user-object
USER_FIELDS = [
    'id',
    'screen_name',
    'name',
    'description',
    'url',
    # 'lang',                               # dropped from tweet objects only by Twitter (since May 19)
    'timestamp_utc',
    'local_time',
    # 'utc_offset',                         # dropped by Twitter (since May 18), ignorable
    # 'time_zone',                          # dropped by Twitter (since May 18), ignorable
    'location',
    # 'geo_enabled',                        # dropped by Twitter (since May 19), ignorable
    'verified',
    'protected',
    'tweets',
    'followers',
    'friends',
    'likes',
    'lists',
    # 'is_translator',                      # dropped by Twitter (since May 19), ignorable
    # 'translator_type',                    # dropped by Twitter (since May 19), ignorable
    # 'is_translation_enabled',             # dropped by Twitter (since May 19), ignorable
    # 'default_profile',
    # 'default_profile_image',
    # 'has_extended_profile',               # dropped by Twitter (since May 19), ignorable
    # 'profile_image_url',                  # dropped by Twitter (since May 19), ignorable
    'image',
    # 'profile_banner_url',
    # 'profile_use_background_image',       # dropped by Twitter (since May 19), ignorable
    # 'profile_background_image_url',       # dropped by Twitter (since May 19), ignorable
    # 'profile_background_image_url_https', # dropped by Twitter (since May 19), ignorable
    # 'profile_background_tile',            # dropped by Twitter (since May 19), ignorable
    # 'profile_background_color',           # dropped by Twitter (since May 19), ignorable
    # 'profile_link_color',                 # dropped by Twitter (since May 19), ignorable
    # 'profile_text_color',                 # dropped by Twitter (since May 19), ignorable
    # 'profile_sidebar_fill_color',         # dropped by Twitter (since May 19), ignorable
    # 'profile_sidebar_border_color'        # dropped by Twitter (since May 19), ignorable
    'default_profile',
    'default_profile_image',
    'witheld_in_countries',
    'witheld_scope'
]

USER_PLURAL_FIELDS = {
    'witheld_in_countries'
}

USER_BOOLEAN_FIELDS = {
    'verified',
    'protected',
    'default_profile',
    'default_profile_image'
}

CANONICAL_URL_KWARGS = {
    'strip_authentication': False,
    'strip_trailing_slash': False,
    'strip_protocol': False,
    'strip_irrelevant_subdomains': False,
    'strip_fragment': False,
    'normalize_amp': False,
    'fix_common_mistakes': False,
    'infer_redirection': False,
    'quoted': True
}

CANONICAL_HOSTNAME_KWARGS = {
    'normalize_amp': False,
    'infer_redirection': False
}

# API v2 constants
TWEET_FIELDS_V2 = {
    'attachments',
    'author_id',
    'context_annotations',
    'conversation_id',
    'created_at',
    'entities',
    'geo',
    'id',
    'in_reply_to_user_id',
    'lang',
    'possibly_sensitive',
    'public_metrics',
    'referenced_tweets',
    'reply_settings',
    'source',
    'text',
    'withheld'
}

MEDIA_FIELDS = {
    'media_key',
    'type',
    'duration_ms',
    'height',
    'preview_image_url',
    'public_metrics',
    'width',
    'alt_text'
}

POLL_FIELDS = {
    'id',
    'options',
    'duration_minutes',
    'end_datetime',
    'voting_status'
}

PLACE_FIELDS = {
    'full_name',
    'id',
    'contained_within',
    'country',
    'country_code',
    'geo',
    'name',
    'place_type'
}

USER_FIELDS_V2 = {
    'id',
    'name',
    'username',
    'created_at',
    'description',
    'entities',
    'location',
    'pinned_tweet_id',
    'profile_image_url',
    'protected',
    'public_metrics',
    'url',
    'verified',
    'withheld'
}

TWEET_EXPANSIONS = {
    'author_id',
    'referenced_tweets.id',
    'in_reply_to_user_id',
    'attachments.media_keys',
    'attachments.poll_ids',
    'geo.place_id',
    'entities.mentions.username',
    'referenced_tweets.id.author_id'
}

TWEET_PARAMS = {
    'tweet.fields': ','.join(TWEET_FIELDS_V2),
    'media.fields': ','.join(MEDIA_FIELDS),
    'poll.fields': ','.join(POLL_FIELDS),
    'place.fields': ','.join(PLACE_FIELDS),
    'user.fields': ','.join(USER_FIELDS_V2)
}

USER_EXPANSIONS = {
    'pinned_tweet_id'
}

USER_PARAMS = {
    'user.fields': ','.join(USER_FIELDS_V2),
    'tweet.fields': ','.join(TWEET_FIELDS_V2)
}

# Lists

LIST_FIELDS = {
    'created_at',
    'follower_count',
    'member_count',
    'private',
    'description',
    'owner_id'
}

LIST_EXPANSIONS = {
    'owner_id'
}

LIST_PARAMS = {
    'list.fields': ','.join(LIST_FIELDS),
    'user.fields': ','.join(USER_FIELDS_V2)
}

LIST_TWEETS_EXPANSIONS = {
    'author_id'
}

LIST_MEMBERS_EXPANSIONS = {
    'pinned_tweet_id'
}

LIST_TWEETS_OR_MEMBERS_PARAMS = {
    'tweet.fields': ','.join(TWEET_FIELDS_V2),
    'user.fields': ','.join(USER_FIELDS_V2)
}

APP_ONLY_ROUTES = {
    'tweets/counts/recent',
    'tweets/counts/all',
    'tweets/search/all'
}



# =============================================================================
# Twitwi Client Wrapper
# =============================================================================
#
# Wrapper for the `twitter` library API client able to rotate the two possible
# endpoints to maximize throughput.
#
import json
from time import sleep, time
from operator import itemgetter
from twitter import Twitter, OAuth, OAuth2, TwitterHTTPError, Twitter2

from twitwi.exceptions import TwitterWrapperMaxAttemptsExceeded
from twitwi.constants import APP_ONLY_ROUTES

DEFAULT_MAX_ATTEMPTS = 5

# Established from: https://developer.twitter.com/en/support/twitter-api/error-troubleshooting
NO_RETRY_STATUSES = set([
    400,
    401,
    403,
    404,
    406,
    410,
    422
])


class TwitterWrapper(object):

    def __init__(self, token, token_secret, consumer_key, consumer_secret,
                 listener=None, api_version='1.1'):

        api_version = str(api_version)

        if api_version not in ['1.1', '2']:
            raise TypeError('API version can only be \'1.1\' or \'2\'.')

        self.oauth = OAuth(
            token,
            token_secret,
            consumer_key,
            consumer_secret
        )

        bearer_token_client = Twitter(
            api_version=None,
            format='',
            secure=True,
            auth=OAuth2(consumer_key, consumer_secret)
        )

        bearer_token = json.loads(
            bearer_token_client.oauth2.token(grant_type='client_credentials')
        )['access_token']

        self.oauth2 = OAuth2(bearer_token=bearer_token)

        self.auth = {}
        self.waits = {}

        TwitterClass = Twitter

        if api_version == '2':
            TwitterClass = Twitter2

            for route in APP_ONLY_ROUTES:
                self.auth[route] = 'app'
                self.waits[route] = {'app': 0}

        self.endpoints = {
            'user': TwitterClass(auth=self.oauth),
            'app': TwitterClass(auth=self.oauth2)
        }

        self.listener = listener

    def call(self, route, max_attempts=DEFAULT_MAX_ATTEMPTS, **kwargs):
        attempts = 0

        if not isinstance(route, list):
            raise TypeError('twitwi.TwitterWrapper.call: expecting route as a list, such as ["friends", "ids"].')

        route = '/'.join(route)

        while attempts < max_attempts:

            if route not in self.auth:
                self.auth[route] = 'user'

            auth = self.auth[route]

            try:
                return self.endpoints[auth].__getattr__(route)(**kwargs)

            except TwitterHTTPError as e:

                # Rate limited
                if e.e.code == 429:
                    now = int(time())

                    # If there are still API calls available, we obviously
                    # queried Twitter too fast and should just let it breathe
                    remaining = int(e.e.headers['X-Rate-Limit-Remaining'])
                    if remaining > 0:
                        attempts += 1

                        if callable(self.listener):
                            self.listener('excessive-rate', {
                                'error': e,
                                'route': route,
                                'attempts': attempts,
                                'auth': auth
                            })

                        sleep(1)
                        continue

                    reset = int(e.e.headers['X-Rate-Limit-Reset'])

                    if route not in self.waits:
                        self.waits[route] = {'user': now, 'app': now}

                    self.waits[route][auth] = reset

                    if callable(self.listener):
                        self.listener('rate-limited', {
                            'route': route,
                            'kwargs': kwargs,
                            'reset': reset,
                            'auth': auth
                        })

                    min_wait = min(self.waits[route].items(), key=itemgetter(1))

                    if min_wait[1] > now:
                        sleeptime = 5 + max(0, int(min_wait[1] - now))

                        if callable(self.listener):
                            self.listener('waiting', {
                                'auth': min_wait[0],
                                'reset': min_wait[1],
                                'sleep': sleeptime
                            })

                        sleep(sleeptime)

                    self.auth[route] = min_wait[0]

                    continue

                # Errors that should terminate immediately
                elif e.e.code in NO_RETRY_STATUSES:
                    raise e

                # Different error
                else:
                    attempts += 1

                    if callable(self.listener):
                        self.listener('error', {
                            'error': e
                        })

        raise TwitterWrapperMaxAttemptsExceeded



# =============================================================================
# Twitwi Library Endpoint
# =============================================================================
#
from twitwi.client_wrapper import TwitterWrapper
from twitwi.formatters import (
    transform_tweet_into_csv_dict,
    format_tweet_as_csv_row,
    transform_user_into_csv_dict,
    format_user_as_csv_row,
    apply_tcat_format
)
# NOTE: should we drop this from public exports?
from twitwi.utils import (
    get_dates,
    custom_normalize_url
)
from twitwi.normalizers import (
    normalize_tweet,
    normalize_user,
    normalize_tweets_payload_v2
)



# =============================================================================
# Twitwi Utilities
# =============================================================================
#
# Miscellaneous utility functions.
#
from pytz import timezone
from ural import normalize_url, get_normalized_hostname
from functools import partial
from datetime import datetime

from twitwi.constants import (
    TWEET_DATETIME_FORMAT,
    FORMATTED_TWEET_DATETIME_FORMAT,
    TWEET_DATETIME_FORMAT_V2,
    CANONICAL_URL_KWARGS,
    CANONICAL_HOSTNAME_KWARGS
)

UTC_TIMEZONE = timezone('UTC')

custom_normalize_url = partial(
    normalize_url,
    **CANONICAL_URL_KWARGS
)

custom_get_normalized_hostname = partial(
    get_normalized_hostname,
    **CANONICAL_HOSTNAME_KWARGS
)


def get_dates(date_str, locale=None, v2=False):
    if locale is None:
        locale = UTC_TIMEZONE

    parsed_datetime = datetime.strptime(date_str, TWEET_DATETIME_FORMAT_V2 if v2 else TWEET_DATETIME_FORMAT)
    utc_datetime = UTC_TIMEZONE.localize(parsed_datetime)
    locale_datetime = utc_datetime.astimezone(locale)

    return (
        int(utc_datetime.timestamp()),
        datetime.strftime(locale_datetime, FORMATTED_TWEET_DATETIME_FORMAT)
    )


def validate_payload_v2(payload):
    if not isinstance(payload, dict):
        return False

    if 'data' not in payload:
        if 'meta' in payload and 'result_count' in payload['meta'] and payload['meta']['result_count'] == 0:
            return True
        else:
            return False

    if not isinstance(payload['data'], list):
        return False

    # NOTE: not sure it cannot be absent altogether
    if 'includes' not in payload or not isinstance(payload['includes'], dict):
        return False

    return True



# =============================================================================
# Twitwi Formatters
# =============================================================================
#
# Miscellaneous formatter functions transforming tweets from one format
# to the other.
#
from twitwi.constants import (
    TWEET_FIELDS,
    GAZOU_TO_TCAT,
    TWEET_BOOLEAN_FIELDS,
    TWEET_PLURAL_FIELDS,
    USER_FIELDS,
    USER_BOOLEAN_FIELDS,
    USER_PLURAL_FIELDS
)


def apply_tcat_format(item):
    result = {v: item[k] for k, v in GAZOU_TO_TCAT['identical_fields'].items() if k in item}
    result['source'] = "<a href=""{}"" rel=""nofollow"">{}</a>".format(item["source_url"], item["source_name"])
    return result


def make_transform_into_csv_dict(plural_fields, boolean_fields):

    def transform_into_csv_dict(item, item_id=None, plural_separator='|', allow_erroneous_plurals=False):
        if item_id is not None:
            item['id'] = item_id

        item['links'] = item.get('proper_links', item.get('links', []))

        for plural_field in plural_fields:
            plurals = item.get(plural_field, [])
            if allow_erroneous_plurals:
                plurals = [element if element is not None else '' for element in plurals]
            item[plural_field] = plural_separator.join(plurals)

        for boolean_field in boolean_fields:
            item[boolean_field] = int(item[boolean_field]) if boolean_field in item else ''

    return transform_into_csv_dict


def make_format_as_csv_row(fields, plural_fields, boolean_fields):

    def format_field_for_csv(field, item, item_id=None, plural_separator='|'):

        if field == 'id' and item_id is not None:
            return item_id

        # NOTE: this is hardly the most performant way to proceed by I am lazy...
        if field in plural_fields:
            v = item.get(field, [])

            if field == 'links':
                v = item.get('proper_links', v)

            return plural_separator.join(v)

        if field in boolean_fields:
            return int(item[field]) if field in item else ''

        return item.get(field, '')

    def format_item_as_csv_row(item, item_id=None, plural_separator='|'):
        return [
            format_field_for_csv(
                field,
                item,
                item_id=item_id,
                plural_separator=plural_separator
            )
            for field in fields
        ]

    return format_item_as_csv_row


transform_tweet_into_csv_dict = make_transform_into_csv_dict(
    TWEET_PLURAL_FIELDS,
    TWEET_BOOLEAN_FIELDS
)

format_tweet_as_csv_row = make_format_as_csv_row(
    TWEET_FIELDS,
    TWEET_PLURAL_FIELDS,
    TWEET_BOOLEAN_FIELDS
)

transform_user_into_csv_dict = make_transform_into_csv_dict(
    USER_PLURAL_FIELDS,
    USER_BOOLEAN_FIELDS
)

format_user_as_csv_row = make_format_as_csv_row(
    USER_FIELDS,
    USER_PLURAL_FIELDS,
    USER_BOOLEAN_FIELDS
)

__all__ = [
    'transform_tweet_into_csv_dict',
    'format_tweet_as_csv_row',
    'transform_user_into_csv_dict',
    'format_user_as_csv_row',
    'apply_tcat_format'
]



from setuptools import setup, find_packages

with open('./README.md', 'r') as f:
    long_description = f.read()

setup(name='twitwi',
      version='0.13.1',
      description='A collection of Twitter-related helper functions for python.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/medialab/twitwi',
      license='MIT',
      author='Béatrice Mazoyer, Guillaume Plique, Benjamin Ooghe-Tabanou',
      author_email='guillaume.plique@sciencespo.fr',
      keywords='twitter',
      python_requires='>=3.4',
      packages=find_packages(exclude=['scripts', 'test']),
      package_data={'docs': ['README.md']},
      install_requires=[
        'pytz>=2019.3',
        'twitter==2.0a0',
        'ural>=0.31.1'
      ],
      zip_safe=True)
